from pydantic import BaseModel, create_model
from pydantic.fields import ModelField
from typing import (
    Any, MutableMapping, Callable, Dict, Optional, List, Type, Union
)
from bson import ObjectId
from functools import lru_cache, partial


def take_last_value(key: str, target: Any, *sources: Any) -> Any:
    """Conflict resolver returning last source
    Args:
        target : current value
        sources : list of update values
    Returns:
        Any: the last value from sources
    """
    try:
        return sources[-1]
    except KeyError:
        return target


def take_first_value(key: str, target: Any, *sources: Any) -> Any:
    """Conflict resolver returning first source
    Args:
        target : current value
        sources : list of update values
    Returns:
        Any: the last value from sources
    """
    try:
        return sources[0]
    except KeyError:
        return target


def merge_values(key: str, target: Any, *sources: Any) -> Any:
    if not isinstance(target, dict) and not key.startswith('$'):
        target = {'$eq': target}
    for s in sources:
        if isinstance(s, MutableMapping):
            target.update(s)
    return target


def dict_deep_update(target: MutableMapping,
                     *sources: MutableMapping,
                     on_conflict: Callable[..., Any] = take_last_value,
                     ) -> MutableMapping:
    """Update a dict with values from others
    Args:
        target : dict to update. (will be modified)
        sources : source dict(s) to update target.
        on_conflict (Callable[[target, *sources], Any], optional):
            define comportment when a key is present in the target
            and/or in multiple source. Defaults to take_last_value.
    Returns:
        MutableMapping: return target so it can be used like this :
        new_dict = dict_deep_update({"key1":"value"}, other)
    """
    merge_keys = []
    for source in sources:
        for k, val in source.items():
            if k in target:
                merge_keys.append(k)
            else:
                target[k] = val
    for k in merge_keys:
        val = target[k]
        src_values = tuple(src[k] for src in sources if k in src)
        if isinstance(val, MutableMapping) and \
                all(isinstance(e, MutableMapping) for e in src_values):
            target[k] = dict_deep_update(val, *src_values,
                                         on_conflict=on_conflict)
        else:
            target[k] = on_conflict(k, val, *src_values)
    return target


def deep_update_model(
    model: BaseModel,
    data: Optional[Dict],
    reset_with_none: bool = True
) -> BaseModel:
    """Update the given model with `data` paylaod (dict) merging childs
    to allow a partial update without loading default values for missing fields

    eratas: since this is a recursive function, this will not work with
    circular dependencies and can raise a RecursionError.
    """
    if data is None:
        return model
    for field, value in data.items():
        node = getattr(model, field, None)
        is_nested_document = isinstance(node, BaseModel)
        if is_nested_document:
            # if the user pass a None to a nested attribute, we want to reset
            # the node.
            if value is None and reset_with_none:
                setattr(model, field, {})
            else:
                deep_update_model(node, data[field])
        else:
            setattr(model, field, value)
    return model


def safe_issubclass(a, b) -> bool:
    try:
        return issubclass(a, b)
    except TypeError:
        return False


def get_all_fields_names(
    model: BaseModel,
    prefix: str = '',
    separator: str = '__',
    field_skip_func: Optional[Callable[[str, ModelField], bool]] = None
) -> List[str]:
    fields = []
    for field_name, field in model.__fields__.items():
        if field_skip_func and field_skip_func(field_name, field):
            continue
        if safe_issubclass(field.type_, BaseModel):
            fields.extend(get_all_fields_names(
                field.type_,
                f'{prefix}{field_name}{separator}',
                separator,
                field_skip_func=field_skip_func
            ))
        else:
            fields.append(prefix + field_name)
    return fields


def get_all_fields(
    model: BaseModel,
    is_ignored: Optional[Callable[[BaseModel, ModelField], bool]] = None,
    node_factory: Type = dict
) -> Dict[str, ModelField]:
    fields = {}
    for field_name, field in model.__fields__.items():
        if is_ignored and is_ignored(model, field):
            continue
        if safe_issubclass(field.type_, BaseModel):
            fields[field_name] = get_all_fields(
                model=field.type_,
                is_ignored=is_ignored
            )
        else:
            fields[field_name] = field
    return node_factory(fields)


def dynamic_model_node_factory(
    model: BaseModel,
    node_data: Dict,
    annotate_all_optional: bool = False
) -> BaseModel:
    def select_annoted_type_from_field(field: Union[ModelField, BaseModel]) -> Type:
        if not isinstance(field, ModelField):
            return field if not annotate_all_optional else Optional[field]
        field_type = model.__annotations__.get(field.name, field.type_)
        return field_type if not annotate_all_optional else Optional[field_type]

    annotations = dict({
        field_name: select_annoted_type_from_field(field)
        for field_name, field in node_data.items()
    })
    optdict = {'__annotations__': annotations}
    optdict.update({
        field.name: field.field_info
        for field in node_data.values()
        if isinstance(field, ModelField)
    })

    class Config:
        json_encoders = {ObjectId: lambda x: str(x)}

    optdict['Config'] = Config
    model = type('DynamicModel', (BaseModel,), optdict)
    return model


def model_map(
    model: BaseModel,
    func: Callable[[BaseModel, ModelField], Optional[Any]],
    node_factory: Callable[[BaseModel, Any, bool], Optional[Any]] = lambda model, data, _: data,
    annotate_all_optional: bool = False
):
    output = {}
    fields: List[Union[ModelField, BaseModel]] = model.__fields__.values()
    for field in fields:
        field = func(model, field)
        if field is None:
            continue
        if safe_issubclass(field.type_, BaseModel):
            output[field.name] = model_map(
                field.type_,
                func,
                node_factory,
                annotate_all_optional
            )
        else:
            output[field.name] = field
    return node_factory(model, output, annotate_all_optional)


@lru_cache
def partial_model(
    baseclass: Type[BaseModel],
    field_filter: Callable[[ModelField], bool] | None = None,
    suffix: str = 'Partial'
) -> Type[BaseModel]:
    """Make all fields in supplied Pydantic BaseModel Optional, for use in PATCH calls.

    Iterate over fields of baseclass, descend into sub-classes, convert fields to Optional and return new model.
    Cache newly created model with lru_cache to ensure it's only created once.
    Use with Body to generate the partial model on the fly, in the PATCH path operation function.

    - https://stackoverflow.com/questions/75167317/make-pydantic-basemodel-fields-optional-including-sub-models-for-patch
    - https://stackoverflow.com/questions/67699451/make-every-fields-as-optional-with-pydantic
    - https://github.com/pydantic/pydantic/discussions/3089
    - https://fastapi.tiangolo.com/tutorial/body-updates/#partial-updates-with-patch
    """
    fields = {}
    for name, field in baseclass.__fields__.items():
        if field_filter and not field_filter(field):
            continue
        type_ = field.type_
        if type_.__base__ is BaseModel:
            fields[name] = (Optional[partial(type_)], {})
        else:
            fields[name] = (
                (Optional[type_], None)
                if field.required
                else (type_, field.default)
            )
    # https://docs.pydantic.dev/usage/models/#dynamic-model-creation
    validators = {"__validators__": baseclass.__validators__}
    return create_model(
        baseclass.__name__ + suffix, **fields, __validators__=validators
    )


def field_mark_filter(mark_list: List[str], field: ModelField) -> bool:
    """Allow filtering of partial models to exclude fields by their names
    if True the field will be kept, otherwise it will be removed from the
    generated partial model.

    ex:
    ```python
    PartialModel = partial_model(partial(field_mark_filter, ['private']))
    ```
    """
    for mark in mark_list:
        if field.field_info.extra.get(mark):
            return False
    return True


def partial_update(
        baseclass: Type[BaseModel],
        suffix='PartialUpdate'
    ) -> Type[BaseModel]:
    """Return a model based on baseclass to partially update it.
    all private & read_only fields will be removed.
    all remaining fields will be optional
    """
    return partial_model(
        baseclass,
        partial(field_mark_filter, ['private', 'read_only']),
        suffix=suffix
    )



import asyncio

from motorized.migration import list_migrations
from depsolve import walk


async def migrate(*folders) -> None:
    """Newer version of migrate for python3.11 using TaskGroup
    need more testing
    """
    migrations = []
    tasks = []
    async with asyncio.TaskGroup() as tg:
        for folder in folders:
            tasks.append(tg.create_task(list_migrations(folder)))

    migrations = [task.result() for task in tasks]

    async with asyncio.TaskGroup() as tg:
        for migration in walk(migrations):
            tg.create_task(migration.apply())



from functools import wraps
from fastapi import status, Query
from fastapi.routing import APIRouter
from fastapi.exceptions import HTTPException

from pydantic import BaseModel
from pydantic.types import NonNegativeInt
from motorized import QuerySet, Document
from motorized.types import InputObjectId
from typing import (
    Type, Literal, List, Optional, Tuple, Callable
)


Action = Literal["create", "list", "delete", "patch", "put", "retrive"]
Method = Literal["GET", "POST", "DELETE", "OPTIONS", "PATCH", "PUT"]


def action(
    path: str,
    method: Method = 'GET',
    status_code: int = status.HTTP_200_OK,
    many=False,
    priority=0
):
    """Register the given function to the availables actions for this view

    Note this decorator only works for `async` views

    `Many` param will only be used if the annoation return is missing in
    addition with the View.get_response_model function
    """
    def decorator(func):
        func._is_action = True
        func._many = many
        func._priority = priority
        # thoses parameters will be passed directly to the APIRouter
        func._router_params = {
            'path': path,
            'methods': [method],
            'status_code': status_code,
        }

        @wraps(func)
        async def wrapper(*args, **kwargs):
            return await func(*args, **kwargs)
        return wrapper
    return decorator


class RestApiView:
    queryset: QuerySet
    # will be used in last resort when resolving return type of an action.
    default_response_model: BaseModel = None
    router: APIRouter
    orderings: List[str] = None

    def __init__(self, router: APIRouter):
        if not hasattr(self, 'queryset'):
            raise ValueError('You MUST define a queryset attribute for views')
        self.router = router
        # populate public ordering if not set.
        if self.orderings is None:
            self.orderings = self.model.get_public_ordering_fields()

        # by default the response model is a model who include all fields
        # except the private ones.
        if not self.default_response_model:
            self.default_response_model = self.model.get_reader_model()
        self._patch_annotations()

    @property
    def model(self) -> Type[Document]:
        return self.queryset.model

    def get_actions_methods(self) -> List[Tuple[str, Callable]]:
        actions_tuples = list([
            (method_name, getattr(self, method_name))
            for method_name in dir(self)
            if self.is_action(method_name)
        ])
        actions_tuples.sort(key=lambda item: item[1]._priority, reverse=True)
        return actions_tuples

    def is_action(self, method_name: str) -> bool:
        try:
            return getattr(self, method_name)._is_action is True
        except AttributeError:
            return False

    def register(self) -> None:
        for method_name, method in self.get_actions_methods():
            params = method._router_params

            # resolve the response model to use.
            try:
                response_model = method.__annotations__['return']
            except KeyError:
                response_model = self.get_response_model(method_name)
                if method._many:
                    response_model = List[response_model]

            self.router.add_api_route(
                **params,
                endpoint=method,
                response_model=response_model
            )

    def get_response_model(self, action: Action) -> Type[BaseModel]:
        """Returns response model for the given action,
        you can override this method if you create new actions or if you want
        to customise the response model (ex: hide fields to user)

        note the prefered way is declaring the annotaion for return of the
        function that you are implementing.
        """
        try:
            return getattr(self, action).__annotations__['return']
        except KeyError:
            if action == 'delete':
                return None
            return self.default_response_model

    def is_implemented(self, action: Action) -> bool:
        try:
            return callable(getattr(self, action))
        except AttributeError:
            return False

    def _patch_annotations(self):
        self.list.__annotations__.update({
            'order_by': Optional[List[self.orderings]],
        })
        updater = self.model.get_updater_model()
        # set the `payload` parameter type
        for action in ('create', 'patch'):
            if not hasattr(self, action):
                continue
            annotations = getattr(self, action).__annotations__
            annotations.setdefault('payload', updater)


class ReadOnlyApiViewSet(RestApiView):
    """Only allow get methods on /resource and /resource/id
    """
    @action('', 'GET', many=True)
    async def list(
        self,
        order_by = Query(None),
        skip: Optional[NonNegativeInt] = Query(None),
        limit: Optional[NonNegativeInt] = Query(10, minimum=1, maximum=50)
    ):
        return await self.queryset.order_by(order_by).skip(skip) \
            .limit(limit).all()

    @action('/{id}', 'GET', priority=-1)
    async def retrieve(self, id: InputObjectId):
        try:
            return await self.queryset.get(_id=id)
        except self.model.DocumentNotFound:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)


class GenericApiView(ReadOnlyApiViewSet):
    @action('', 'POST', status.HTTP_201_CREATED)
    async def create(self, payload):
        instance = self.model(**payload.dict())
        return await instance.commit()

    @action('/{id}', 'DELETE', priority=-1)
    async def delete(self, id: InputObjectId):
        await self.queryset.filter(_id=id).delete()

    @action('/{id}', 'PATCH', priority=-1)
    async def patch(self, id: InputObjectId, payload):
        try:
            instance = await self.queryset.get(_id=id)
            instance.deep_update(payload.dict(exclude_unset=True))
            await instance.save()
            return instance
        except self.model.DocumentNotFound:
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)



class Criteria:
    def __init__(self, value=None):
        self.value = value

    def __hash__(self):
        return hash(self.name)

    def command(self, invert=False) -> str:
        raise NotImplementedError('override this')

    def get_value(self, invert=False):
        return self.value

    def as_mongo_expression(self, invert=False):
        return {self.command(invert): self.get_value(invert)}


class Eq(Criteria):
    def command(self, invert=False):
        return '$eq' if not invert else '$ne'


class Neq(Criteria):
    def command(self, invert=False):
        return '$ne' if not invert else '$eq'


class In(Criteria):
    def command(self, invert=False):
        return '$in' if not invert else '$nin'


class Nin(Criteria):
    def command(self, invert=False):
        return '$nin' if not invert else '$in'


class Gt(Criteria):
    def command(self, invert=False):
        return '$gt' if not invert else '$lte'


class Lt(Criteria):
    def command(self, invert=False):
        return '$lt' if not invert else '$gte'


class Lte(Criteria):
    def command(self, invert=False):
        return '$lte' if not invert else '$gt'


class Gte(Criteria):
    def command(self, invert=False):
        return '$gte' if not invert else '$lt'


class Exists(Criteria):
    def command(self, invert=False):
        return '$exists'

    def get_value(self, invert=False):
        if invert:
            return self.value is False
        return self.value


class Regex(Criteria):
    def command(self, invert=False):
        return '$regex'

    def get_value(self, invert=False):
        if not invert:
            return self.value
        raise NotImplementedError('Inverting regex is not implemented yet')



from .client import connection  # noqa: F401
from .document import (
    Document,
    BaseModel,
    EmbeddedDocument,
    PrivatesAttrsMixin,
    Field,
    mark_parents
)  # noqa: F401
from .query import Q  # noqa: F401
from .queryset import QuerySet  # noqa: F401




from inspect import isclass
from typing import Optional, Union, Any, Dict, Type, List, Generator, Literal, Tuple
from pydantic import BaseModel, Field, validate_model
from pydantic.fields import ModelField
from pydantic.main import ModelMetaclass
from pymongo.results import InsertOneResult, UpdateResult

from motorized.queryset import QuerySet
from motorized.query import Q
from motorized.types import PydanticObjectId, ObjectId
from motorized.exceptions import DocumentNotSavedError, MotorizedError
from motorized.utils import (
    deep_update_model, get_all_fields_names,
    model_map, dynamic_model_node_factory
)


class DocumentMeta(ModelMetaclass):
    def __new__(cls, name, bases, optdict: Dict) -> Type['Document']:
        # remove any reference to `objects` in the class
        # we only declare it to be readable and conveniant with the IDE
        # the filtering of objects is only effective if you declare a class
        # to allow to give values or a real field using this name.
        try:
            objects = optdict['__annotations__']['objects']
            if isclass(objects):
                optdict['__annotations__'].pop('objects')
                optdict.pop('objects', None)
        except KeyError:
            pass

        # we allocate the BaseModel with pydantic metaclass
        instance: Type[Document] = super().__new__(cls, name, bases, optdict)
        if name not in ('Document',):
            cls._populate_default_mongo_options(
                cls, name, instance, optdict.get('Mongo'))

        class DocumentError(MotorizedError):
            pass

        class TooManyMatchException(DocumentError):
            pass

        class DocumentNotFound(DocumentError):
            pass

        instance.DocumentError = DocumentError
        instance.TooManyMatchException = TooManyMatchException
        instance.DocumentNotFound = DocumentNotFound
        instance.objects = instance.Mongo.manager_class(
            instance, getattr(instance.Mongo, 'filters', None))
        return instance

    def _populate_default_mongo_options(cls, name: str, instance: "Document",
                                        custom_mongo_settings_class) -> None:
        class Mongo:
            pass

        # forbid re-utilisation of the Mongo class between
        # inheritance of the class except for the `manager_class` parameter
        previous_manager_class = None
        try:
            if instance.Mongo.class_name != name:
                instance.Mongo = Mongo()
                previous_manager_class = instance.Mongo.manager_class
        except AttributeError:
            pass

        if custom_mongo_settings_class:
            instance.Mongo = custom_mongo_settings_class

        default_settings = {
            'collection': name.lower() + 's',
            'manager_class': previous_manager_class or QuerySet,
            'local_fields': [],
            'class_name': name,
            'filters': None,
        }

        for attribute_name, default_value in default_settings.items():
            if not hasattr(instance.Mongo, attribute_name):
                setattr(instance.Mongo, attribute_name, default_value)


class DocumentBasis(BaseModel):
    """Represent the very bassis of Document and EmbeddedDocument
    """
    class Config:
        json_encoders = {ObjectId: str}
        validate_assignment = True

    def update(self, input_data: Dict) -> "Document":
        """Update the current instance with the given `input_data` after
        validation return the object itself (without saving it in the database)
        """
        validate_model(self, input_data)
        allow_extra: bool = getattr(self.Config, 'extra', 'ignore') == 'allow'

        # load the fields into the current instance
        for field, value in input_data.items():
            if allow_extra or hasattr(self, field):
                setattr(self, field, value)
        return self

    def deep_update(self, input_data: Dict, **kwargs) -> "Document":
        return deep_update_model(self, input_data, **kwargs)

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith('_'):
            return object.__setattr__(self, name, value)

        return super().__setattr__(name, value)


class PrivatesAttrsMixin:
    """Ommit any private arributes from iterators and .dict method on this
    document.
    """
    class Config:
        json_encoders = {ObjectId: str}
        validate_assignment = True
        underscore_attrs_are_private = True

    def __iter__(self) -> Generator[str, Any, None]:
        for field_name, field_value in super().__iter__():
            if field_name not in self.__fields__:
                continue
            yield field_name, field_value

    def dict(self, **kwargs) -> Dict[str, Any]:
        def resolve_field(value):
            if isinstance(value, list):
                return list([resolve_field(item) for item in value])
            if isinstance(value, dict):
                return dict({k: resolve_field(v) for k, v in value.items()})
            if isinstance(value, BaseModel):
                return value.dict(**kwargs)
            return value

        return dict({
            field_name: resolve_field(field_value)
            for field_name, field_value in self
        })


class EmbeddedDocument(DocumentBasis):
    pass


class Document(DocumentBasis, metaclass=DocumentMeta):
    objects: QuerySet
    id: Optional[PydanticObjectId] = Field(alias='_id', read_only=True)

    class Mongo:
        manager_class = QuerySet

    def __init__(self, *args, **kwargs) -> None:
        BaseModel.__init__(self, *args, **self._transform(**kwargs))

    def get_query(self) -> Q:
        document_id = getattr(self, 'id', None)
        if not document_id:
            raise DocumentNotSavedError('document has no id.')
        return Q(_id=document_id)

    async def _create_in_db(self, creation_dict: Dict) -> InsertOneResult:
        response = await self.objects.insert_one(creation_dict)
        self.id = response.inserted_id
        return response

    async def _update_in_db(self, update_dict: Dict) -> UpdateResult:
        return await self.objects.collection.update_one(
            filter={'_id': self.id},
            update={'$set': update_dict}
        )

    @classmethod
    def _is_field_to_save(cls, field_name: str) -> bool:
        return (
            not field_name.startswith('_') and
            field_name not in cls.Mongo.local_fields and
            field_name in cls.__fields__
        )

    async def to_mongo(self) -> Dict:
        """Convert the current model dictionary to database output dict,
        this also mean the aliased fields will be stored in the alias name
        instead of their name in the document declaration.
        """
        saving_data = self.dict()

        # remove any field that is not to save, this has to be done
        # before the aliasing resolving to allow to save/load fields
        # that starts with _
        saving_data = dict(
            {
                k: v for k, v in saving_data.items()
                if self._is_field_to_save(k)
            }
        )

        # resolve all alised fields to be saved in their alias name
        for field in self._aliased_fields():
            saving_data[field.alias] = saving_data.pop(field.name, None)

        return saving_data

    async def save(
        self,
        force_insert: bool = False
    ) -> Union[InsertOneResult, UpdateResult]:
        data = await self.to_mongo()
        document_id = data.pop('_id', None)
        if document_id is None or force_insert:
            if force_insert:
                data['_id'] = document_id
            return await self._create_in_db(data)
        return await self._update_in_db(data)

    async def commit(self) -> "Document":
        """Same as `.save` but return the current instance.
        """
        await self.save()
        return self

    async def delete(self) -> "Document":
        """Delete the current instance from the database,
        to the deleted the instance need to have a .id set, in any case the
        function will return the instance itself
        """
        try:
            await self.objects.filter(self.get_query()).delete_one()
        except DocumentNotSavedError:
            pass
        setattr(self, 'id', None)
        return self

    async def fetch(self) -> "Document":
        """Return a fresh instance of the current document from the database.
        """
        return await self.objects.filter(self.get_query()).get()

    @classmethod
    def _aliased_fields(cls) -> Generator[List[ModelField], None, None]:
        """Return the list of fields with aliases
        """
        return [
            field for field in cls.__fields__.values()
            if field.name != field.alias
        ]

    def _transform(self, **kwargs) -> Dict:
        """Override this method to change the input database before having it
        being validated/parsed by BaseModel (pydantic)
        """
        return kwargs

    async def reload(self) -> "Document":
        # fetch an validate input data from database
        model_data = await self.objects.filter(self.get_query()).find_one()
        model_data.pop('_id')
        return self.update(model_data)

    def __repr__(self):
        def get_field_entry(field: Field) -> str:
            return f"{field.name}={getattr(self, field.name)}"

        fields = ", ".join([
            get_field_entry(field)
            for field in self.__fields__.values()
            if field.name not in self.Mongo.local_fields
        ])
        return f'{self.__class__.__name__}({fields})'

    # def __setattr__(self, name: str, value: Any) -> None:
    #     # allow any private attribute to be passed
    #     if name.startswith('_') or name in self.Mongo.local_fields:
    #         return object.__setattr__(self, name, value)

    #     # otherwise we let pydantic decide
    #     return super().__setattr__(name, value)

    @classmethod
    def get_readonly_fields(cls):
        return list(cls.get_marked_fields('read_only').keys())

    @classmethod
    def get_marked_fields(
            cls, mark: str,
            value=True,
            default=False
    ) -> Dict[str, ModelField]:
        return dict({
            field_name: field
            for field_name, field in cls.__fields__.items()
            if field.field_info.extra.get(mark, default) == value
        })

    @classmethod
    def get_updater_model(
        cls,
        exclude: Optional[List[Tuple[BaseModel, Optional[List[str]]]]] = None
    ) -> Type[BaseModel]:
        return cls.get_filtered_model(exclude, ['read_only', 'private'])

    @classmethod
    def get_reader_model(
        cls,
        exclude: Optional[List[Tuple[BaseModel, Optional[List[str]]]]] = None,
        exclude_fields_marks: Optional[List[str]] = None
    ) -> Type[BaseModel]:
        if exclude_fields_marks is None:
            exclude_fields_marks = []
        reader = cls.get_filtered_model(exclude, ['private'] + exclude_fields_marks)
        model = type(cls.__name__ + 'Reader', (reader,), {})
        return model

    @classmethod
    def get_filtered_model(
        cls,
        exclude: Optional[List[Tuple[BaseModel, Optional[List[str]]]]] = None,
        exclude_fields_marks: Optional[List[str]] = None
    ) -> Type[BaseModel]:
        """This class factory function create a new BaseModel from this model
        with all the fields that are not marked as `read_only`, all the fields
        are optional in the generated model

        exemple:
        ```python
        updater = User.get_updater_model(exclude=[(User, ['password'])])
        ```
        """
        def is_excluded(model: BaseModel, field: ModelField) -> bool:
            if not exclude:
                return False
            for exclude_model, exclude_field_names in exclude.items():
                if model == exclude_model:
                    if exclude_field_names is None:
                        return True
                    if field.field_info.name in exclude_field_names:
                        return True
            return False

        def field_filtering(model: BaseModel,
                            field: ModelField) -> Optional[ModelField]:
            if is_excluded(model, field):
                return None
            if exclude_fields_marks:
                for mark in exclude_fields_marks:
                    if field.field_info.extra.get(mark):
                        return None
            if field.name in cls.Mongo.local_fields:
                return None
            return field

        return model_map(cls, field_filtering, dynamic_model_node_factory, True)

    @classmethod
    def get_public_ordering_fields(cls) -> Literal:
        """Return a literal with all visibles fields from an external use
        perspective (ignore all Fields(private=True))
        """
        def is_ignored(field_name: str, field: ModelField) -> bool:
            if field.field_info.extra.get('private', False):
                return True
            if field_name in cls.Mongo.local_fields:
                return True
            return False

        fields = get_all_fields_names(cls, field_skip_func=is_ignored)
        fields.extend([f'-{field_name}' for field_name in fields])
        fields.sort(key=lambda x: x if not x.startswith('-') else x[1:])
        return Literal[tuple(fields)]


def mark_parents(
    model: DocumentBasis,
    parent: Optional[DocumentBasis] = None
) -> None:
    """Mark all nested items with a `_parent` attribute pointing to their
    parent instance (for trees of DocumentBasis)

    it is strongly recomemded to use this on classes with the
    `PrivatesAttrsMixin` mixin to avoid ciruclar loop when calling the
    .dict method.
    """
    model._parent = parent
    for field in model.__fields__.values():
        field: ModelField = field
        try:
            if not issubclass(field.type_, DocumentBasis):
                continue
        # in this error all the Literal will fall, we don't want to mark them
        # so we just continue.
        except TypeError:
            continue

        item = getattr(model, field.name)
        if isinstance(item, DocumentBasis):
            mark_parents(item, model)

        elif isinstance(item, list):
            for submodel in item:
                mark_parents(submodel, model)

        elif isinstance(item, dict):
            for submodel in item.values():
                mark_parents(submodel, model)



from typing import List, Any, Dict, Tuple
from motorized.keywords import (
    Eq, Neq, In, Nin, Gte, Lte, Gt, Lt, Exists, Regex
)
from motorized.utils import merge_values, dict_deep_update


KEYWORDS = {
    'eq': Eq,
    'neq': Neq,
    'in': In,
    'nin': Nin,
    'gte': Gte,
    'ge': Gte,
    'lte': Lte,
    'le': Lte,
    'gt': Gt,
    'lt': Lt,
    'exists': Exists,
    'regex': Regex
}


def dict_path(path: List[str], value: Any = None) -> Dict:
    """Construct a dictionary from a path to hold the given value,
    example:
    d = QuerySet.dict_path(['a', 'b', 'c'], 42)
    d == {'a': {'b': {'c': 42}}}
    """
    out = {}
    node = out
    last_node = None
    last_key = None
    for k in path:
        node[k] = {}
        last_node = node
        node = node[k]
        last_key = k
    last_node[last_key] = value
    return out


class QueryDict(dict):
    def __init__(self, **kwargs) -> None:
        data = {}
        for key, value in kwargs.items():
            path = key.split('__')
            node = dict_path(path, value)
            dict_deep_update(data, node, on_conflict=merge_values)
        super().__init__(**data)


class Q:
    def __init__(self, **kwargs) -> None:
        self.query = self.convert_kwargs_to_query(**kwargs)

    def __repr__(self) -> str:
        return f'<Q: {self.query}>'

    def copy(self) -> "Q":
        instance = Q()
        instance.query = self.query.copy()
        return instance

    @classmethod
    def raw(cls, query: Dict) -> "Q":
        if not isinstance(query, dict):
            raise TypeError(f'A dictionary was expected, got {type(query)}')
        instance = cls()
        instance.query = query
        return instance

    @classmethod
    def convert_kwargs_to_query(cls, invert=False, **kwargs) -> Dict:
        query = {}
        for key, value in kwargs.items():
            path, value = cls.apply_keywords(
                value,
                key.split('__'),
                invert=invert
            )
            filter_dict = dict_path(path, value)
            dict_deep_update(query, filter_dict,
                             on_conflict=merge_values)
        return query

    @staticmethod
    def read_dict_path(data: dict, path: List['str']) -> Any:
        x = data
        for node in path:
            x = x[node]
        return x

    @classmethod
    def apply_keywords(
        cls,
        raw_value: Any,
        path: List[str],
        invert: bool = False
    ) -> Tuple[str, Any]:
        if invert and path[-1] not in KEYWORDS:
            path.append('eq')

        for cmd in path:
            try:
                op = KEYWORDS[cmd](raw_value)
                return path[0:-1], op.as_mongo_expression(invert)
            except KeyError:
                pass
        return path, raw_value

    def __add__(self, other: "Q") -> "Q":
        instance = self.copy()
        dict_deep_update(instance.query, other.query, on_conflict=merge_values)
        return instance

    def is_empty(self) -> bool:
        return not self.query

    def __eq__(self, other: "Q") -> bool:
        return self.query == other.query

    def __or__(self, other: "Q") -> "Q":
        return Q.raw({'$or': [self.query, other.query]})

    def __and__(self, other: "Q") -> "Q":
        return Q.raw({'$and': [self.query, other.query]})



import os
from glob import glob
from typing import Literal, Callable, Any, Optional, Dict, Type
from datetime import datetime
from importlib import import_module
from motorized import Document
import logging
import asyncio
from depsolve import walk


"""
# Migrations
each migration is supposed be a file in a folder (of your choice),
each file nameing should be: YYYYMMDDXX
when:
YYYY = year on 4 chars
MM = month on 2 chars
DD = day on 2 chars
XX = migration of the day from 00 to 99 (on 2 chars)

all migrations will be evaluated and applied in a lexical order.

each migration should contain at last:
- async apply function to perform the migration
- async revert function to revert it if possible

it's also a good practice to provide a `description` variable on the top scope
of your migration module to let co-workers know what it is intended for.

example migration of a file "2021021400.py":
```python
from motorized.migration import alter_field
from my_models import User

description = 'convert User.age from string to integer'


async def apply() -> int:
    return alter_field(User, 'age', lambda age: int(age)})


async def revert() -> int:
    return alter_field(User, 'age', lambda age: str(age))
```
"""


logger = logging.getLogger(__name__)

# thoses are the types from
# https://docs.mongodb.com/manual/reference/bson-types/
MongoType = Literal[
    "double",
    "string",
    "object",
    "array",
    "binData"
    "undefined",
    "objectId",
    "bool",
    "date",
    "null",
    "regex",
    "dbPointer",
    "javascript",
    "symbol",
    "javascriptWithScope",
    "int",
    "timestamp",
    "long",
    "decimal",
    "minKey",
    "maxKey",
    "binData",
]


class Migration(Document):
    module_name: str
    applied_at: Optional[datetime] = None
    depends_on = []

    class Mongo:
        local_fields = ('depends_on',)

    @property
    def name(self) -> str:
        return self.module_name

    def __str__(self):
        return self.module_name

    async def is_applied(self) -> bool:
        if self.id and self.applied_at:
            return True
        return await Migration.objects \
            .filter(module_name=self.module_name).exists()

    @property
    def path(self) -> str:
        return self.module_name.replace('.', '/') + '.py'

    @property
    def exists(self) -> bool:
        return os.path.exists(self.path)

    async def save(self, *args, **kwargs):
        if not self.applied_at:
            self.applied_at = datetime.utcnow()
        return await super().save(*args, **kwargs)

    async def apply(self, force: bool = False) -> int:
        """Apply the migration to the database and save the instance into
        the collection.
        `force` param allow you to force application of the migration even if
        it was already applied.
        if the migration has already been applied then 0 will be returned
        and a warning displayed
        return the amount of modified rows in the collection
        """
        if await self.is_applied() and not force:
            logger.info(f"{self.module_name} already applied.")
            return 0

        migration_module = import_module(self.module_name)
        if not hasattr(migration_module, 'apply'):
            logger.error(f"{self} is malformed: no `apply` method found.")
            raise ValueError(self)

        # description formating if available
        description = getattr(migration_module, 'description', None)
        description: str = ': ' + description if description else '.'

        # perform the actual migration
        modified_count = await migration_module.apply()
        logger.info(f"Applied {self.module_name} on {modified_count} rows{description}")
        await self.save()
        return modified_count

    async def revert(self) -> int:
        if not await self.is_applied():
            logger.warning(f"Cannot revert {self.module_name} since it wasent applied")
            return 0
        migration_module = import_module(self.module_name)
        try:
            modified_count = await migration_module.revert()
            print('revert function ok', modified_count)
            logger.info(f"Reverted {self.module_name} on {modified_count} rows.")
            await self.delete()
            return modified_count
        except AttributeError as error:
            logger.error("Migration {self.module_name} cannot be undone.")
            raise ValueError(f"{self} cannot be undone.") from error

    def __eq__(self, other: "Migration") -> bool:
        return self.module_name == other.module_name

    @classmethod
    def from_module(cls, module: str) -> "Migration":
        migration_module = import_module(module)
        if not hasattr(migration_module, 'apply'):
            raise ValueError(f"Module {module} does not have a apply function")
        migration = cls(
            module_name=module,
            depends_on=getattr(migration_module, 'depends_on', []),
            applied_at=None
        )
        return migration


def value_from_dot_notation(data: Dict[str, Any], path: str) -> Any:
    """Take a dictionary `data` and get keys by the `path` parameter,
    this path parameter will use the dots (.) as delimiter.
    """
    for key in path.split('.'):
        data = data[key]
    return data


async def alter_field(
    model: Type[Document],
    field_name: str,
    caster: Callable[[Any], Any],
    filter: Optional[Dict[str, Any]] = None,
) -> int:
    """Alter the `field_name` on `model` using the `caster` function,
    the function will receive the value to convert as parameter.

    the alteration will be done row by row instead of a collection.update_many
    to allow custom conversions (cast/setting default values etc).
    """
    if filter is None:
        filter = {}
    projection = {field_name: True, '_id': True}
    cursor = model.collection.find(filter, projection)
    modified_count = 0
    async for item in cursor:
        casted = await caster(value_from_dot_notation(item, field_name))
        await model.collection.update_one(
            {'_id': item['_id']},
            {'$set': {field_name: casted}}
        )
        modified_count += 1
    return modified_count


async def list_migrations(folder: str) -> list[Migration]:
    migrations = []
    for item in glob(folder + '/*.py'):
        if not os.path.isfile(item):
            continue
        module_name = (
            item
            .replace('/', '.')
            .replace('.py', '')
        )
        try:
            migrations.append(Migration.from_module(module_name))
        except ValueError:
            continue
    return migrations


async def migrate(*folders: str) -> None:
    migrations = []
    for folder in folders:
        migrations.extend(await list_migrations(folder))

    for migrations in walk(migrations):
        tasks = list([migration.apply() for migration in migrations])
        await asyncio.gather(*tasks)
